/******************************************************************************
 *                                                                            *
 * File:   gsm0338 .c         Version  1.00            Date: 2001-03-30       *
 *                                                                            *
 ******************************************************************************
 *                                                                            *
 * Function:                                                                  *
 *                                                                            *
 * filter converting GMS 03.38 7-bit character code to UTF-8/UCS2             *
 *                                                                            *
 * Src: GSM 03.38 (common for SMS)                                            *
 *      V1.00 by Kosta Kostis <kosta@kostis.net>                              *
 *      Source: http://www.dreamfabric.com/sms/default_alphabet.html          *
 *                                                                            *
 * Dst: ISO/IEC 10646-1:2000 aka Unicode 3.0.1                                *
 *      V3.0.1 by Unicode Consortium http://www.unicode.org/                  *
 *      Source: ftp://ftp.unicode.org/Public/UNIDATA/UnicodeData.txt          *
 *      converted, sorted & edited by Kosta Kostis <kosta@kostis.net>         *
 *                                                                            *
 ******************************************************************************
 *                                                                            *
 * History:                                                                   *
 *     2001-03-30: transtab V2.00                                             *
 *        - initial generation                                                *
 *                                                                            *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef	_WIN32
	#include <io.h>
	#include <fcntl.h>
#endif

#include "gsm0338.h"
#include "gsm0338.tab"
#include "utf8.h"

/*****************************************************************************/

static	int	UCS2		= 0 ;	/* UTF-8 or UCS2 output format */
static	int	Unbuffered	= 0 ;	/* to buffer or not to buffer */
static	int	Reverse		= 0 ;	/* reverse conversion */

/******************************************************************************

	Function:
		display copyright, author and program information

	Parameters:
		none

	Returns:
		nothing

 *****************************************************************************/

static  void    Hello (void)
{
	fprintf (stderr, MSG_COPYRIGHT) ;
	fprintf (stderr, MSG_VERSION, PRGNAM, VERSION) ;
	fprintf (stderr, MSG_FROM_TO,
		SRC_NAME, SRC_COMMENT1, SRC_COMMENT2, SRC_COMMENT3,
		DST_NAME, DST_COMMENT1, DST_COMMENT2, DST_COMMENT3) ;
}

/******************************************************************************

	Function:
		swap endian

	Parameters:
		unsigned short          value to be swapped

	Returns:
		unsigned short		swapped value

 *****************************************************************************/

static  unsigned short	swap_endian
(
	unsigned short value
)
{
	int	lsb, msb ;

	lsb = value & 0xFF ;
	msb = (value & 0xFF00) >> 8 ;

	return ((lsb << 8) | msb) ;
}

/******************************************************************************

	Function:
		convert UCS2 to GSM 03.38 and display

	Parameters:
		FILE    *Dst    	destination file
		unsigned short	utf8	Unicode character

	Returns:
		ERR_NONE                if all goes well
		ERR_BAD                 if characters have been untranslatable

 *****************************************************************************/

int	gsm0338_out
(
	FILE	*Dst,
	unsigned short	utf8
)
{
	int	i ;

	/* check for regular characters, first */
	for (i = 0 ; i < CHARNUM ; ++i)
	{
		if (TRANS_C (i) == utf8)
		{
			if (fputc (i, Dst) != i)
				return (ERR_BAD) ;
			return (ERR_NONE) ;
		}
	}

	/* check for escape characters, second */
	for (i = 0 ; i < CHARNUM ; ++i)
	{
		if (TRANS_ESC_C (i) == utf8)
		{
			if (fputc (ESCAPE, Dst) != ESCAPE)
				return (ERR_BAD) ;
			if (fputc (i, Dst) != i)
				return (ERR_BAD) ;
			return (ERR_NONE) ;
		}
	}

	return (ERR_BAD) ;
}

/******************************************************************************

	Function:
		copy input to output doing reverse conversion

	Parameters:
		FILE    *Src            source file
		FILE    *Dst            destination file

	Returns:
		ERR_NONE                if all goes well
		ERR_BAD                 if characters have been untranslatable

 *****************************************************************************/

static	int	RevTrans
(
	FILE	*Src,
	FILE	*Dst
)
{
	unsigned short ucs2 ;
	unsigned short c ;
	uchar32	utf8 ;
	int	codes	= ERR_NONE ;
	int	swap	= 2 ;

	if (UCS2)
	{
		c = (unsigned short) UCS2_INTRO ;
		fread (&ucs2, sizeof (ucs2), 1, Src) ; 
		if (ucs2 == c)
			swap = 0 ;
		if (ucs2 == swap_endian (c))
			swap = 1 ;
		if (swap == 2)
			return (ERR_BAD) ;
	}

	do
	{
		if (UCS2)
		{
			if (fread (&ucs2, sizeof (ucs2), 1, Src) != 1)
				return (ERR_BAD) ;
			if (swap)
				ucs2 = swap_endian (ucs2) ;
			utf8 = (uchar32) ucs2 ;
		}
		else
		{
			if (UTF8_Read (Src, &utf8) != UTF8_OK)
				return (ERR_BAD) ;
		}
		if (feof (Src) == 0)
			gsm0338_out (Dst, (unsigned short) utf8) ;
	}
	while (feof (Src) == 0) ;

	return (ERR_NONE) ;
}

/******************************************************************************

	Function:
		copy input to output doing conversion

	Parameters:
		FILE    *Src            source file
		FILE    *Dst            destination file

	Returns:
		ERR_NONE                if all goes well
		ERR_BAD                 if characters have been untranslatable

 *****************************************************************************/

static  int     Trans
(
	FILE    *Src,
	FILE    *Dst
)
{
	int     ch ;			/* character read               */
	uchar32	newch ;			/* translated character         */
	uchar32	escch ;			/* ESCAPE character		*/
	unsigned short c ;
	int     codes	= ERR_NONE ;	/* bad code flag                */
	int	rc ;

	if (UCS2)
	{
		c = (unsigned short) UCS2_INTRO ;
		fwrite (&c, sizeof (c), 1, Dst) ; 
	}

	while ((ch = fgetc (Src)) != EOF)	/* read character       */
	{
		newch = (uchar32) TRANS_C (ch) ;
		if (newch != (int) BAD)
		{
			if (newch == ESCAPE)	/* ESCAPE */
			{
				ch = fgetc (Src) ;
				if (ch == EOF)
					continue ;
				escch = (uchar32) TRANS_ESC_C (ch) ;
				if (escch != (int) BAD)
				{
					if (UCS2 == 0)
					{
						rc = UTF8_Write (Dst, escch) ;
						if (rc != UTF8_OK)
							return (ERR_BAD) ;
					}
					else
					{
						c = (unsigned short) escch ;
						rc = fwrite (&c, sizeof (c), 1, Dst) ; 
						if (rc != 1)
							return (ERR_BAD) ;
					}
				}
				else
				{
					codes = ERR_BAD ;
				}
			}
			else
			{
				if (UCS2 == 0)
				{
					rc = UTF8_Write (Dst, newch) ;
					if (rc != UTF8_OK)
						return (ERR_BAD) ;
				}
				else
				{
					c = (unsigned short) newch ;
					rc = fwrite (&c, sizeof (c), 1, Dst) ;
					if (rc != 1)
						return (ERR_BAD) ;
				}
			}
		}
		else
		{
			codes = ERR_BAD ;
		}
	}

	return (codes) ;
}


/******************************************************************************

	ScanFlags () return codes

 *****************************************************************************/

#define	SCN_OK		0	/* parameters are all valid                  */
#define	SCN_UNKNOWN	1	/* unknown parameter used                    */
#define	SCN_BADARGS	2	/* illegal number of params                  */

/******************************************************************************

	Function:
		parse parameters and set flags and variables accordingly

	Parameters:
		int     argc    number of parameters
		char    *argv[] argument strings

	Returns:
		SCN_OK          if all parameters are valid
		SCN_UNKNOWN     if any of the parameters is unknown
		SCN_BADARGS     if the number of parameters is wrong

 *****************************************************************************/

static  int     ScanFlags
(
	int	argc,
	char	*argv	[]
)
{
	int	i ;		/* index through the parameters              */
	int	parms = 0 ;	/* yet no parameter scanned                  */

	/**********************************************************************

		set defaults / initialize

	 *********************************************************************/

	for (i = 1 ; i < argc ; ++i)    /* scan command line                 */
	{
		/**************************************************************

			is it a flag/switch? (*NIX style)

		 *************************************************************/

		if (*argv [i] == '-')
		{
			/******************************************************

				check out which flag it is

			 *****************************************************/

			switch (argv [i][1])
			{
				case	'u':	/* change BAD character      */
				case	'U':

					UCS2 = 1 ;
					break ;

				case	'n':	/* unbuffered I/O */
				case	'N':

					Unbuffered = 1 ;
					break ;

				case	'r':	/* reversed conversion */
				case	'R':

					Reverse = 1 ;
					break ;

				default:        /* Uh! This flag's unknown!  */

					return (SCN_UNKNOWN) ;
			}
		}
		else
		{
			/******************************************************

				if it's no flag it's bad

			 *****************************************************/


			return (SCN_BADARGS) ;
		}
	}

	/**********************************************************************

		everything's been fine - let's quit!

	 *********************************************************************/

	return (SCN_OK) ;
}

/*****************************************************************************/

int     main
(
	int     argc,
	char    *argv   []
)
{
	FILE    *SRCfile = stdin ;      /* Source default is stdin           */
	FILE    *DSTfile = stdout ;     /* Dest. default is stdout           */
	char    *Help ;
	int     codes ;                 /* bad code flag                     */
	int     syntax ;                /* return code for ScanFlags ()      */

	/**********************************************************************

		parse parameters/arguments

	 *********************************************************************/

	syntax = ScanFlags (argc, argv) ;
	if (syntax != SCN_OK)
	{
		/**************************************************************

			display syntax error message and quit

		 *************************************************************/

		Hello () ;
		fprintf (stderr, ERR_MSG_SYNTAX, PRGNAM) ;

		return (ERR_SYNTAX) ;
	}

	/**********************************************************************

		set streams to binary mode

	 *********************************************************************/

#ifdef	_WIN32

	_setmode (fileno (stdin), _O_BINARY) ;
	_setmode (fileno (stdout), _O_BINARY) ;

#endif

	/**********************************************************************

		set streams to unbuffered mode if selected

	 *********************************************************************/

	if (Unbuffered)
	{
		setvbuf (SRCfile, NULL, _IONBF, 0);
		setvbuf (DSTfile, NULL, _IONBF, 0);
	}

	/**********************************************************************

		now do the actual conversion

	 *********************************************************************/

	if (Reverse == 0)
		codes = Trans (SRCfile, DSTfile) ;
	else
		codes = RevTrans (SRCfile, DSTfile) ;

	return (codes) ;
}
