/******************************************************************************
 *                                                                            *
 * File:   utf8.c             Version  1.31            Date: 2001-03-05       *
 *                                                                            *
 * Copyright (C) 1993-2001 by Kostis Netzwerkberatung                         *
 * Talstr. 25, D-63322 Rdermark, Tel. +49 6074 881056, FAX 881058            *
 * kosta@kostis.net (Kosta Kostis), http://www.kostis.net/                    *
 *                                                                            *
 *         This program may be used free of charge at your own risk.          *
 *                                                                            *
 ******************************************************************************
 *                                                                            *
 * History:                                                                   *
 *     2001-03-05: KK V1.31                                                   *
 *        - added exceptions for u[0]=F8|FC                                   *
 *     2001-03-04: KK V1.30                                                   *
 *        - added support for UTF-32                                          *
 *        - rewrote code to work in loops (more general)                      *
 *     2001-03-02: KK V1.22                                                   *
 *        - cosmetic changes                                                  *
 *     2001-02-25: KK V1.21                                                   *
 *        - improved error handling reading UTF-8, changed error codes        *
 *     2001-02-24: KK V1.20                                                   *
 *        - implemented suggestions provided by Florian Weimer                *
 *        - corrected bug for values between U+0080 and U+07FF (gee!)         *
 *        - changed interface from strings to files                           *
 *     2001-02-23: KK V1.11                                                   *
 *        - checking for surrogates (U+D800-U+DFFF, U+FFFE, U+FFFF)           *
 *     2001-02-17: KK V1.10                                                   *
 *        - http://www.unicode.org/unicode/uni2errata/UTF-8_Corrigendum.html  *
 *          now checking for non-shortest form more thoroughly                *
 *          this version is limited to single UTF-16 charaters (<=0xFFFF)     *
 *     1994-05-05: KK V1.01                                                   *
 *        - some compilers still don't understand "//" comments               *
 *     1993-07-09: KK V1.00                                                   *
 *        - initial coding                                                    *
 *                                                                            *
 *****************************************************************************/

#include <stdio.h>

#include "utf8.h"

/******************************************************************************

	Function:
		check if character is surrogate or illegal

	Parameters:
		uchar32	ch		UTF-32 character to be checked

	Returns:
		UTF8_OK			if UTF-32 is not surrogate or illegal
		UTF8_SURROGATE		Unicode surrogate detected
		UTF8_ILLEGAL		illegal Unicode detected

 *****************************************************************************/

int	UTF8_Check
(
	uchar32		ch
)
{
	/*********************************************************************

		surrogates are codes U+D800..U+DFFF

	 ********************************************************************/

	if ((ch >= 0xD800) && (ch <= 0xDFFF))
		return (UTF8_SURROGATE) ;

	/*********************************************************************

		illegal codes are U+FFFE, U+FFFF and all code >= 0x80000000

	 ********************************************************************/

	if
	(
		(ch == ISO10646_BAD) ||
		(ch == ISO10646_BAD2) ||
		(ch >= 0x80000000)
	)
		return (UTF8_ILLEGAL) ;

	return (UTF8_OK) ;
}

/******************************************************************************

	Function:
		 write UTF-8 for a given UTF-32 Unicode character

	Parameters:
		FILE	*UTF8		UTF-8 file being read
		uchar32	ch		UTF-32 character to be written

	Returns:
		UTF8_OK			if all went well
		UTF8_NULLPOINTER	NULL pointer detected
		UTF8_SURROGATE		Unicode surrogate detected
		UTF8_ILLEGAL		illegal Unicode detected
		UTF8_EOF		error while writing UTF-8 string

 *****************************************************************************/

int	UTF8_Write
(
	FILE	*FUTF8,
	uchar32	ch
)
{
	int	rc ;
	int	octets ;
	int	i ;
	int	mask ;
	int	shift ;
	int	c ;

	/**********************************************************************

		check for NULL pointer

	 *********************************************************************/

	if (FUTF8 == (FILE *) NULL)
		return (UTF8_NULLPOINTER) ;

	/**********************************************************************

		check for surrogates and illegal codes

	 *********************************************************************/

	if ((rc = UTF8_Check (ch)) != UTF8_OK)
		return (rc) ;

	/**********************************************************************

		check for transparent us-ascii

	 *********************************************************************/

	if (ch <= 0x7F)
	{
		if (fputc ((int) ch, FUTF8) == EOF)
			return (UTF8_EOF) ;

		return (UTF8_OK) ;
	}

	/**********************************************************************

		calculate number of (additional) octets to be written

	 *********************************************************************/

	octets = 1 ;		/* 1 octet  for     U+0080..    U+07FF */
	if (ch >= 0x0800)
		++octets ;	/* 2 octets for     U+0800..    U+FFFF */
	if (ch >= 0x10000)
		++octets ;	/* 3 octets for    U+10000..  U+1FFFFF */
	if (ch >= 0x200000)
		++octets ;	/* 4 octets for   U+200000.. U+3FFFFFF */
	if (ch >= 0x4000000)
		++octets ;	/* 5 octets for  U+4000000..U+7FFFFFFF */

	/**********************************************************************

		write UTF-8 string (first octet is special)

	 *********************************************************************/

	mask = (1 << (6 - octets)) - 1 ; /* 0x1F,0x0F,0x07,0x03,0x01 */
	shift = octets * 6 ;		/* 30,24,18,12,6 */
	c = 0xFF - ((mask << 1) + 1) ;	/* can you follow this? ;) */
	c += (int) ((ch & (mask << shift)) >> shift) ;
	if (fputc (c, FUTF8) == EOF)
		return (UTF8_EOF) ;

	/**********************************************************************

		write rest of UTF-8 string

	 *********************************************************************/

	shift -= 6 ;
	mask = 0x3F << shift ;
	for (i = 1 ; i <= octets ; ++i, shift -= 6)
	{
		c = (int) (ch & mask) >> shift ;
		mask = mask >> 6 ;
		if (fputc (0x80 | (c & 0x3F), FUTF8) == EOF)
			return (UTF8_EOF) ;
	}

	return (UTF8_OK) ;
}

/******************************************************************************

	Function:
		read UTF-8 character, store as UTF-32 Unicode

	Parameters:
		FILE	*FUTF8		UTF-8 file to be read
		uchar32	*ch		UTF-32 Unicode

	Returns:
		UTF8_OK			if all goes well
		UTF8_NULLPOINTER	NULL pointer detected
		UTF8_MALFORMED		if UTF-8 syntax is not met
		UTF8_SURROGATE		Unicode surrogate detected
		UTF8_ILLEGAL		illegal Unicode detected
		UTF8_EOF		unexpected EOF encountered

 *****************************************************************************/

int	UTF8_Read
(
	FILE	*FUTF8,
	uchar32	*ch
)
{
	int	u	[6] ;	/* max. # of UTF-8 octets */
	int	octets ;
	int	i ;
	int	mask ;
	int	shift ;

	/**********************************************************************

		check for NULL pointers

	 *********************************************************************/

	if ((FUTF8 == (FILE *) NULL) || (ch == (uchar32 *) NULL))
		return (UTF8_NULLPOINTER) ;

	/**********************************************************************

		initialize return value (assume the worst ;) )

	 *********************************************************************/

	*ch = ISO10646_BAD ;

	/**********************************************************************

		read first octet from file (checking for EOF)

	 *********************************************************************/

	if ((u[0] = fgetc (FUTF8)) == EOF)
		return (UTF8_EOF) ;
	
	/**********************************************************************

		check for values <= U+007F

	 *********************************************************************/

	if (u[0] <= 0x7F)
	{
		/**************************************************************

			such values are encoded transparently (us-ascii)

		 *************************************************************/

		*ch = (uchar32) u[0] ;

		return (UTF8_OK) ;
	}
	
	/**********************************************************************

		u[0] values from 0x80 to 0xC1 and 0xFE to 0xFF are illegal

	 *********************************************************************/

	if ((u[0] < 0xC2) || (u[0] > 0xFD))
		return (UTF8_MALFORMED) ;

	/**********************************************************************

		calculate number of (additional) octets to be read

	 *********************************************************************/

	octets = 1 ;		/* 1 octet  for     U+0080..    U+07FF */
	if (u[0] > 0xDF)
		++octets ;	/* 2 octets for     U+0800..    U+FFFF */
	if (u[0] > 0xEF)
		++octets ;	/* 3 octets for    U+10000..  U+1FFFFF */
	if (u[0] > 0xF7)
		++octets ;	/* 4 octets for   U+200000.. U+3FFFFFF */
	if (u[0] > 0xFB)
		++octets ;	/* 5 octets for  U+4000000..U+7FFFFFFF */

	/**********************************************************************

		read (additional) octets from file (checking for EOF)

	 *********************************************************************/

	for (i = 1 ; i <= octets ; ++i)
		if ((u[i] = fgetc (FUTF8)) == EOF)
			return (UTF8_EOF) ;

	/**********************************************************************

		check validity of octets read

	 *********************************************************************/

	for (i = 1 ; i <= octets ; ++i)
		if ((u[i] < 0x80) || (u[i] > 0xBF))
			return (UTF8_MALFORMED) ;

	/**********************************************************************

		special exceptions

	 *********************************************************************/

	if ((u[0] == 0xE0) && (u[1] < 0xA0))
		return (UTF8_MALFORMED) ;
	if ((u[0] == 0xF0) && (u[1] < 0x90))
		return (UTF8_MALFORMED) ;
	if ((u[0] == 0xF8) && (u[1] < 0x88))
		return (UTF8_MALFORMED) ;
	if ((u[0] == 0xFC) && (u[1] < 0x84))
		return (UTF8_MALFORMED) ;

	/**********************************************************************

		calculate value

	 *********************************************************************/

	mask = (1 << (6 - octets)) - 1 ; /* 0x1F,0x0F,0x07,0x03,0x01 */
	shift = octets * 6 ;		/* 30,24,18,12,6 */
	*ch = (uchar32) ((u[0] & mask) << shift) ;
	for (i = 1, shift -= 6 ; i <= octets ; ++i, shift -= 6)
		*ch += (uchar32) ((u[i] & 0x3F) << shift) ;

	return (UTF8_Check (*ch)) ;
}

